//=============================================================================
//||                             DirixleJacobi                               ||
//||                                                                         ||
//||                               MainGPU.cu                                ||
//||                                                                         ||
//||                               by Dampir                                 ||
//=============================================================================

#include "CudaHeaders.cuh"

#define BLOCK_SIZE 16
#define CntRunKernel 10


__device__  float MAX_VALUE;

__device__ void warp_reduce_max(volatile float* smem)
{

	smem[threadIdx.x] = smem[threadIdx.x+32] > smem[threadIdx.x] ? 
						smem[threadIdx.x+32] : smem[threadIdx.x]; __syncthreads();

	smem[threadIdx.x] = smem[threadIdx.x+16] > smem[threadIdx.x] ? 
						smem[threadIdx.x+16] : smem[threadIdx.x]; __syncthreads();

	smem[threadIdx.x] = smem[threadIdx.x+8] > smem[threadIdx.x] ? 
						smem[threadIdx.x+8] : smem[threadIdx.x]; __syncthreads();

	smem[threadIdx.x] = smem[threadIdx.x+4] > smem[threadIdx.x] ? 
						smem[threadIdx.x+4] : smem[threadIdx.x]; __syncthreads();

	smem[threadIdx.x] = smem[threadIdx.x+2] > smem[threadIdx.x] ? 
						smem[threadIdx.x+2] : smem[threadIdx.x]; __syncthreads();

	smem[threadIdx.x] = smem[threadIdx.x+1] > smem[threadIdx.x] ? 
						smem[threadIdx.x+1] : smem[threadIdx.x]; __syncthreads();

}

template<int threads>
__global__ void find_min_max_dynamic(float* in, float* out, int n, int start_adr, int num_blocks)
{	
	__shared__ float smem_max[64];

	int tid = threadIdx.x + start_adr;

	float max = -inf;
	float val;

	// tail part
	int mult = 0;
	for(int i = 1; mult + tid < n; i++)
	{
		val = in[tid + mult];	
		
		max = val > max ? val : max;

		mult = int_mult(i,threads);
	}

	// MAX part
	for(int i = 1; mult+threadIdx.x < num_blocks*2; i++)
	{
		val = out[threadIdx.x + mult];

		max = val > max ? val : max;
		
		mult = int_mult(i,threads);
	}

	if(threads == 32)		
		smem_max[threadIdx.x+32] = 0.0f;	
	smem_max[threadIdx.x] = max;

	__syncthreads();

	if(threadIdx.x < 32)		
		warp_reduce_max(smem_max);
	
	if(threadIdx.x == 0) {
		out[blockIdx.x + gridDim.x] = smem_max[threadIdx.x]; //return max
		//out[blockIdx.x + gridDim.x] = smem_max[2]; //return max
		MAX_VALUE = out[2];
	}	
}

template<int els_per_block, int threads>
__global__ void find_min_max(float* in, float* out, int* indices)
{
	__shared__ float smem_max[64];

	int tid = threadIdx.x + blockIdx.x*els_per_block;

	float max = -inf;	
	float val;	
	int max_index = -1;

	const int iters = els_per_block/threads;
#pragma unroll
		for(int i = 0; i < iters; i++)
		{
			val = in[tid + i*threads];						
			if(val > max) {
				max = val;
				max_index = tid+int_mult(i,threads); 
			}			
		}
		
	if(threads == 32)	
		smem_max[threadIdx.x+32] = 0.0f;	
	
	smem_max[threadIdx.x] = max;

	__syncthreads();

	if(threadIdx.x < 32)
		warp_reduce_max(smem_max);	
	
	__syncthreads();

	if(threadIdx.x == 0)		
		out[blockIdx.x + gridDim.x] = smem_max[threadIdx.x]; 		
}

void findBlockSize(int* whichSize, int* num_el)
{
	const float pretty_big_number = 24.0f*1024.0f*1024.0f;

	float ratio = float((*num_el))/pretty_big_number;

	if(ratio > 0.8f)
		(*whichSize) =  5;
	else if(ratio > 0.6f)
		(*whichSize) =  4;
	else if(ratio > 0.4f)
		(*whichSize) =  3;
	else if(ratio > 0.2f)
		(*whichSize) =  2;
	else
		(*whichSize) =  1;
}

void compute_reduction(float* d_in, float* d_out, int* d_indices, int num_els)
{	
	int whichSize = -1; 
		
	findBlockSize(&whichSize,&num_els);

	int block_size = powf(2,whichSize-1)*blockSize1;
	int num_blocks = num_els/block_size;
	int tail = num_els - num_blocks*block_size;
	int start_adr = num_els - tail;
	
	if(whichSize == 1)
		find_min_max<blockSize1,threads><<< num_blocks, threads>>>(d_in, d_out, d_indices); 
	else if(whichSize == 2)
		find_min_max<blockSize1*2,threads><<< num_blocks, threads>>>(d_in, d_out, d_indices); 
	else if(whichSize == 3)
		find_min_max<blockSize1*4,threads><<< num_blocks, threads>>>(d_in, d_out, d_indices); 
	else if(whichSize == 4)
		find_min_max<blockSize1*8,threads><<< num_blocks, threads>>>(d_in, d_out, d_indices); 
	else
		find_min_max<blockSize1*16,threads><<< num_blocks, threads>>>(d_in, d_out, d_indices); 
	
   find_min_max_dynamic<threads><<< 1, threads>>>(d_in, d_out, num_els, start_adr, num_blocks);	

}

__global__ void flt_test_compute_per_block_unew(float *u, float *u_new, float *f, float eps, int n, int m, float hx, float hy)
{
	 int tx    = threadIdx.x;       
    int ty    = threadIdx.y;
	int bx    = blockIdx.x;        
    int by    = blockIdx.y;

	int begin = n*(BLOCK_SIZE-1)*by + (BLOCK_SIZE-1)*bx;
 
	__shared__ float shU_old [BLOCK_SIZE+1][BLOCK_SIZE+1];
	__shared__ float shU_new [BLOCK_SIZE+1][BLOCK_SIZE+1];
	__shared__ float shF     [BLOCK_SIZE+1][BLOCK_SIZE+1];

	shU_old[ty][tx] = u    [begin + n*ty + tx]; 
	shU_new[ty][tx] = u_new[begin + n*ty + tx];
	shF[ty][tx]     = f    [begin + n*ty + tx];
 	__syncthreads();

	if ((ty > 0) && (ty < BLOCK_SIZE) && (tx > 0) && (tx < BLOCK_SIZE))							
 		shU_new[ty][tx] = 0.25f * (shU_old[ty][tx-1] + shU_old[ty][tx+1] + shU_old[ty-1][tx] + shU_old[ty+1][tx] - hx * hy * shF[ty][tx]);

	if ((ty > 0) && (ty < BLOCK_SIZE) && (tx > 0) && (tx < BLOCK_SIZE))
		u_new[begin + n*ty + tx] = shU_new[ty][tx];	
} 	

__global__	void vectors_dif(float *u1, float *u2, float *sum, int length)
{
	int i = blockIdx.x * blockDim.x + threadIdx.x; 
 	int j = blockIdx.y * blockDim.y + threadIdx.y;
	int tid = i*blockDim.y + j;
	sum[tid] = fabs(u1[tid] - u2[tid]); 	
}


void test_compute_per_blocks(float *u, float *u_new, float *u_exact, float *f, fltParamOfMatrix &POM)
{
	FILE *fp;

    float *devU        = NULL;
    float *devU_new    = NULL;    
    float *devF        = NULL; 
    int    n           = POM.N;
	int    m           = POM.M;
    float  hx          = POM.hx;
    float  hy          = POM.hy;
    float  eps         = POM.eps;

	// for get max value 
	//--------------------	
	int size = n*m*sizeof(float);
	int index_size = 4096*sizeof(int); // max 2048 blocks! Bug warning!!
	float *sum;	
	int* d_indices;	
	float* d_out;
	float* h_sum = (float*) malloc(size);
	cudaMalloc((void**)&d_out, size);
	cudaMalloc((void**)&d_indices, index_size);
	cudaMalloc((void**)&sum,   size);    
	//--------------------

	void *d_p = NULL;
	float *h_p = (float*) malloc(sizeof(float));
	cudaGetSymbolAddress(&d_p, MAX_VALUE);
	float *devExact   = NULL;
	float *exactArray = NULL;
		
	int grid_width = 1;
	int grid_heigth = 1;
	int CntOfIter = 0; 
   	float  gpuTime     = 0.f;

   	cudaEvent_t        start, stop;  	

   dim3 threads1(BLOCK_SIZE + 1, BLOCK_SIZE + 1);

   if ((n - 2*BLOCK_SIZE) == 0){
	   grid_width = ceil((float)n/BLOCK_SIZE);
	   grid_heigth = ceil((float)n/BLOCK_SIZE);
   }
   else 
	   if ( (n - 2*BLOCK_SIZE) % BLOCK_SIZE*(BLOCK_SIZE-1) == 0) {
		grid_width = (int) (ceil((float)n/BLOCK_SIZE) + (n - 2*BLOCK_SIZE)/(BLOCK_SIZE*(BLOCK_SIZE-1)));
		grid_heigth = (int) (ceil((float)n/BLOCK_SIZE) + (n - 2*BLOCK_SIZE)/(BLOCK_SIZE*(BLOCK_SIZE-1)));
	   }
   dim3 grid   (grid_width, grid_heigth);  

   
   cudaMalloc((void**)&devU,        size); 
   cudaMalloc((void**)&devU_new,    size);   
   cudaMalloc((void**)&devF,        size);       
   cudaMalloc((void**)&devExact, grid_width*grid_heigth*sizeof(float));
	
   cudaEventCreate(&start);
   cudaEventCreate(&stop );    
   cudaMemcpy(devU,       u,       size, cudaMemcpyHostToDevice); 		
   cudaMemcpy(devU_new,   u_new,   size, cudaMemcpyHostToDevice);	  
   cudaMemcpy(devF,       f,       size, cudaMemcpyHostToDevice); 
   
   cudaEventRecord (start, 0);
   do {
	for (int k = 0; k < CntRunKernel; ++k) {
		flt_test_compute_per_block_unew<<<grid, threads1>>>(devU, devU_new, devF, eps, n, m, hx, hy);			
		flt_test_compute_per_block_unew<<<grid, threads1>>>(devU_new, devU, devF, eps, n, m, hx, hy);  						
	}
	
	vectors_dif <<<grid, threads>>>(devU, devU_new, sum, n*m);	
	compute_reduction(sum, d_out, d_indices, n*m);	
	cudaThreadSynchronize();

	cudaMemcpy(h_sum, d_out, size, cudaMemcpyDeviceToHost);	
	cudaMemcpy(h_p, d_p, sizeof(float), cudaMemcpyDeviceToHost);	

	fp = fopen("Result.txt", "w");
	for (int i = 0; i < n; ++i) {
	   for (int j = 0; j < m; ++j)	
		   fprintf(fp, "%0*.*f\t",1,6, h_sum[n*i + j]);	
	   fprintf(fp, "\n");
	}
   fclose(fp);   
   } while (*h_p > eps);


   cudaEventRecord      (stop, 0);
   cudaEventSynchronize (stop);
   cudaEventElapsedTime (&gpuTime, start, stop );  
	
   cudaMemcpy(u,             devU,        size,    cudaMemcpyDeviceToHost);    

   fp = fopen("Result.txt", "w");
   for (int i = 0; i < n; ++i) {
	   for (int j = 0; j < m; ++j)	
		   fprintf(fp, "%0*.*f\t",1,6, u[n*i + j]);	
	   fprintf(fp, "\n");
   }
   fclose(fp);

   fp = fopen("Iteration_Time.log", "a");
   fprintf(fp, "N\t M\t hx\t\t\t hy\t\t\t\t eps\t\t\t gpuTime\t CntOfIter\t\n", n);
   fprintf(fp, "%d\t ", n);
   fprintf(fp, "%d\t ", m);
   fprintf(fp, "%f\t ", hx);
   fprintf(fp, "%f\t\t ", hy);
   fprintf(fp, "%f\t\t ", eps);
   fprintf(fp, "%f\t\t ", gpuTime);
   fprintf(fp, "%d\n\n", CntOfIter);
   fclose(fp);

   cudaFree(devF       ); 
   cudaFree(devU       );
   cudaFree(devU_new   );
   cudaFree(sum);
   cudaFree(devExact);
   cudaFree(d_out);
   cudaFree(d_indices);
   free(exactArray);
   cudaEventDestroy (start);
   cudaEventDestroy (stop );
   return;
}

